import os
import re

from pathlib import Path
from typing import Literal

from .logs import logstr
from .decorations import brp
from .matches import logger, iterate_folder
from .types import PathType, StrsType
from .paths import norm_path
from .confirms import confirm_input


def content_match(content: str, pattern: str):
    if not content or not pattern:
        return None
    return re.search(pattern, content)


def pathname_match(path: PathType, pattern: str):
    if not path or not pattern:
        return None
    return re.search(pattern, str(path))


def basename_match(path: PathType, pattern: str):
    if not path or not pattern:
        return None
    return re.search(pattern, Path(path).name)


def sort_paths_by_depth(paths: list[PathType]) -> list[PathType]:
    """Sort paths by their depth in the file system."""
    if not paths:
        return []
    return sorted(paths, key=lambda x: len(str(x).split("/")), reverse=True)


def read_file_content(path: PathType) -> str:
    """Read the content of a file."""
    if not path:
        return ""
    with open(path, "r", encoding="utf-8") as f:
        return f.read()


def log_rename(
    old: PathType,
    new: PathType,
    text_type: Literal["file", "folder", "content"],
    indent: int = 2,
):
    """Log the renaming of a file or folder."""
    with logger.temp_indent(indent):
        logger.note(f"> Renaming {text_type}:")
        logger.warn(f"  * from: {old}")
        logger.okay(f"  * to  : {new}")


def log_paths_to_rename(paths_to_rename: dict[str, list[PathType]]):
    """Log the paths that will be renamed."""
    logger.note(f"> Following paths will be renamed:")
    for text_type, paths in paths_to_rename.items():
        if not paths:
            continue
        logger.mesg(f"  * {text_type}:")
        for p in paths:
            logger.hint(f"    * {p}")


def log_match(p: PathType, text_type: Literal["content", "file", "folder"]):
    """Log the matching of a file or folder."""
    logger.hint(f"! {p} {logstr.mesg(brp(text_type))}")


def rename_texts(
    root: PathType,
    renames_dict: dict,
    includes: StrsType = None,
    excludes: StrsType = None,
    unmatch_bool: bool = True,
    use_gitignore: bool = True,
    is_rename_content: bool = True,
    is_rename_file: bool = True,
    is_rename_folder: bool = True,
    confirm_before_rename: bool = True,
    verbose: bool = True,
) -> dict[str, list[tuple[PathType, PathType]]]:
    """
    Params:
        - root: root directory to search for files and folders to rename
        - renames_dict: dict of patterns and replacements
            - keys: patterns to match
            - vals: new values to replace the matched patterns with
        - includes: list of patterns to include files/folders
        - excludes: list of patterns to exclude files/folders
        - is_rename_content: whether to rename file contents
        - is_rename_file: whether to rename file names
        - is_rename_folder: whether to rename folder names

    Rename order: content -> file -> folder
        - contents: file paths whose file content hits pattern
        - files: file paths whose basename hit pattern
        - folders: folder paths whose pathname hit pattern
    """
    root = norm_path(root)
    logger.note(f"> Renaming directory:", end=" ")
    logger.file(f"[{root}]")

    paths_to_rename = {"contents": [], "folders": [], "files": []}
    paths_renamed = {"contents": [], "folders": [], "files": []}

    for p, match_bool, level in iterate_folder(
        root,
        includes=includes,
        excludes=excludes,
        yield_folder=True,
        unmatch_bool=unmatch_bool,
        use_gitignore=use_gitignore,
        verbose=verbose,
    ):
        if not match_bool:
            continue

        # as renaming contents would not break file structure, we could do it directly
        # but to be consistent with file and folder renames, we would do it later
        if is_rename_content and p.is_file():
            content = read_file_content(p)
            for pattern, new_val in renames_dict.items():
                if content_match(content, pattern):
                    paths_to_rename["contents"].append(p)
                    log_match(p, "content")
                    break

        # as renaming files would break file structure, we would do it later
        if is_rename_file and p.is_file():
            for pattern, new_val in renames_dict.items():
                if basename_match(p, pattern):
                    paths_to_rename["files"].append(p)
                    log_match(p, "file")
                    break

        # as renaming folders would break file structure, we would do it later
        if is_rename_folder and p.is_dir():
            for pattern, new_val in renames_dict.items():
                if basename_match(p, pattern):
                    paths_to_rename["folders"].append(p)
                    log_match(p, "folder")
                    break

    if confirm_before_rename:
        log_paths_to_rename(paths_to_rename)
        if not confirm_input("rename", op_name="rename texts"):
            logger.warn("× rename_texts cancelled by user")
            return

    # rename file contents
    if is_rename_content and paths_to_rename["contents"]:
        for p in paths_to_rename["contents"]:
            new_content = read_file_content(p)
            for pattern, new_val in renames_dict.items():
                new_content = re.sub(pattern, new_val, new_content)
            log_rename(p, p, "content")
            with open(p, "w", encoding="utf-8") as wf:
                wf.write(new_content)
            paths_renamed["contents"].append((p, p))

    # rename file basenames
    if is_rename_file and paths_to_rename["files"]:
        for p in paths_to_rename["files"]:
            new_basename = Path(p).name
            for pattern, new_val in renames_dict.items():
                new_basename = re.sub(pattern, new_val, new_basename)
            new_path = Path(p).parent / new_basename
            log_rename(p, new_path, "file")
            os.rename(p, new_path)
            paths_renamed["files"].append((p, new_path))

    # rename folder basenames
    if is_rename_folder and paths_to_rename["folders"]:
        # sort folders to rename from the deepest to the shallowest
        paths_to_rename["folders"] = sort_paths_by_depth(paths_to_rename["folders"])
        for p in paths_to_rename["folders"]:
            new_basename = Path(p).name
            for pattern, new_val in renames_dict.items():
                new_basename = re.sub(pattern, new_val, new_basename)
            new_path = Path(p).parent / new_basename
            log_rename(p, new_path, "folder")
            os.rename(p, new_path)
            paths_renamed["folders"].append((p, new_path))

    return paths_renamed



from datetime import datetime, timedelta
from typing import Literal, Union

from .colors import colored
from .logs import logger, add_fills
from .times import get_now, t_to_str, dt_to_str, dt_to_sec


class Runtimer:
    def __init__(self, verbose=True):
        self.verbose = verbose

    def __enter__(self):
        self.start_time()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.end_time()
        self.elapsed_time()

    def start_time(self):
        self.t1 = get_now()
        self.logger_time("start", self.t1)
        return self.t1

    def end_time(self):
        self.t2 = get_now()
        self.logger_time("end", self.t2)
        return self.t2

    def elapsed_time(self):
        self.dt = self.t2 - self.t1
        self.logger_time("elapsed", self.dt)
        return self.dt

    @property
    def elapsed_seconds(self) -> float:
        if hasattr(self, "dt"):
            pass
        else:
            self.dt = self.end_time() - self.start_time()
        return dt_to_sec(self.dt, precision=2)

    def logger_time(
        self,
        time_type: Literal["start", "end", "elapsed"],
        t: Union[datetime, timedelta],
    ):
        if not self.verbose:
            return
        time_types = {
            "start": "Start",
            "end": "End",
            "elapsed": "Elapsed",
        }

        if isinstance(t, datetime):
            t_str = t_to_str(t)
        elif isinstance(t, timedelta):
            t_str = dt_to_str(t)
        else:
            t_str = str(t)

        if time_type == "elapsed":
            time_color = "light_green"
            fill_color = "light_green"
        else:
            time_color = "light_magenta"
            fill_color = "light_magenta"
        time_str = colored(f"{time_types[time_type]} time: [ {t_str} ]", time_color)

        filled_time_str = add_fills(
            time_str,
            filler="* ",
            fill_side="both",
            is_text_colored=True,
            fill_color=fill_color,
        )
        logger.line(filled_time_str)



import ast
import json
import re

from typing import Union, Any

RE_CODE_BLOCK = re.compile(r"```(?:json)?\s*(.*?)\s*```", re.DOTALL | re.IGNORECASE)


class JsonParser:
    """Extract json part from text, which encapsulated by ```json ... ``` or ``` ```."""

    def __init__(self):
        pass

    def parse(self, text: str) -> Any:
        match = RE_CODE_BLOCK.search(text)
        if not match:
            return None
        json_str = match.group(1)
        json_obj = None
        try:
            json_obj = json.loads(json_str)
        except json.JSONDecodeError:
            try:
                json_obj = ast.literal_eval(json_str)
            except (SyntaxError, ValueError):
                pass
        return json_obj

    def parse_all(self, text: str) -> list[Any]:
        res: list[Any] = []
        for match in RE_CODE_BLOCK.finditer(text):
            json_str = match.group(1)
            try:
                json_obj = json.loads(json_str)
                res.append(json_obj)
            except json.JSONDecodeError:
                try:
                    json_obj = ast.literal_eval(json_str)
                    res.append(json_obj)
                except (SyntaxError, ValueError):
                    pass
        return res



from .logs import logger, logstr
from .decorations import brk


def confirm_input(
    expect_val: str,
    op_name: str = "",
    ignore_case: bool = False,
    max_retries: int = None,
    raise_error: bool = True,
    comp_func: callable = None,
    exec_func: callable = None,
    *exec_args,
    **exec_kwargs,
) -> bool:
    comp_res = False
    if not op_name:
        op_name = exec_func.__name__ if exec_func else "operation"
    attempts = 0
    hint_prefix = ""
    while not comp_res:
        if isinstance(max_retries, int):
            hint_prefix = f"[{logstr.file(attempts + 1)}/{logstr.mesg(max_retries)}] "
            if attempts >= max_retries:
                logger.warn(f"× Exceed max_retries ({max_retries}). Stop confirm.")
                break
        input_val = input(
            logstr.note(
                f'> {hint_prefix}Type "{logstr.file(expect_val)}" to confirm {logstr.mesg(brk(op_name))}: '
            )
        )
        if comp_func:
            comp_res = comp_func(input_val, expect_val)
        else:
            if ignore_case:
                comp_res = input_val.lower() == expect_val.lower()
            else:
                comp_res = input_val == expect_val
        attempts += 1

    if comp_res:
        if exec_func:
            exec_func(*exec_args, **exec_kwargs)
        return True
    else:
        if raise_error:
            raise ValueError(f"Confirmation failed")
        return False



from .args import MergedArgParser
from .types import PathType, PathsType, KeyType, KeysType, StrsType, IntsType
from .types import LIST_TYPES, DictListType
from .exceptions import BreakpointException, raise_breakpoint
from .colors import colored, decolored
from .logs import TCLogger, logger, TCLogstr, logstr, log_error
from .fills import add_fills
from .times import get_now, get_now_ts, get_now_str, get_now_ts_str, get_date_str
from .times import TIMEZONE, set_timezone, tcdatetime
from .times import ts_to_str, str_to_ts, str_to_t
from .times import t_to_str, t_to_ts, dt_to_sec, dt_to_str
from .times import unify_ts_and_str
from .runtimes import Runtimer
from .dicts import CaseInsensitiveDict, dict_get, dict_set, dict_get_all, dict_set_all
from .dicts import dict_pop, dict_extract, dict_clean, dict_flatten
from .jsons import JsonParser
from .envs import OSEnver, shell_cmd
from .maths import int_bits, max_key_len, chars_len
from .maths import is_str_float, to_digits, get_by_threshold
from .formats import DictStringifier, dict_to_str
from .tables import is_listable, norm_any_to_str_list, norm_any_to_type_list
from .tables import dict_to_table_str
from .files import FileLogger
from .bars import TCLogbar, TCLogbarGroup
from .decorations import brk, brc, brp
from .strings import chars_slice
from .attrs import attrs_to_dict
from .params import obj_param, obj_params
from .params import obj_params_dict, obj_params_list, obj_params_tuple
from .matches import match_val, match_key, iterate_folder, match_paths
from .confirms import confirm_input
from .paths import norm_path, strf_path
from .copies import copy_file, copy_file_relative, copy_folder
from .trees import tree_folder
from .renames import rename_texts



import argparse


class MergedArgParser:
    """Usage:
    ```python
    arg_parser = MergedArgParser(ParserClass1, ParserClass2, ...)
    arg_parser = add_parser_class(ParserClass3, ...)
    args = arg_parser.parse_args()
    ```
    """

    def __init__(self, *parser_classes):
        self.parser_classes = parser_classes
        self.construct_parsers()

    def construct_parsers(self):
        self.parsers = [cls(add_help=False) for cls in self.parser_classes]

    def add_parser_class(self, *args):
        self.parser_classes += args
        self.construct_parsers()

    def parse_args(self, args=None):
        self.args = argparse.ArgumentParser(parents=self.parsers).parse_args()
        return self.args



"""OS env and shell utils"""

import json
import os
import subprocess

from pathlib import Path

from .colors import colored
from .logs import logger
from .dicts import CaseInsensitiveDict


class OSEnver:
    def __init__(self, secrets_json=None):
        if not secrets_json:
            self.secrets_json = Path(__file__).parent / "secrets.json"
        else:
            self.secrets_json = secrets_json
        self.load_secrets()

    def load_secrets(self):
        self.secrets = CaseInsensitiveDict()
        try:
            with open(self.secrets_json, mode="r", encoding="utf-8") as rf:
                secrets = json.load(rf)
                for key, value in secrets.items():
                    self.secrets[key] = value
        except Exception as e:
            logger.warn(f"Loading local secrets: {e}")

    def __getitem__(self, key=None):
        if key:
            return self.secrets.get(key, os.getenv(key, None))
        else:
            return dict(self.secrets.items())


def shell_cmd(cmd, getoutput=False, showcmd=True, env=None):
    if showcmd:
        logger.info(colored(f"\n$ [{os.getcwd()}]", "light_blue"))
        logger.info(colored(f"  $ {cmd}\n", "light_cyan"))
    if getoutput:
        output = subprocess.getoutput(cmd)
        return output
    else:
        subprocess.run(cmd, shell=True, env=env)



"""Math utils"""

import math
from typing import Union, Literal


def int_bits(num, base: int = 10):
    """Calc bits of integer with base X"""
    return math.ceil(math.log(num + 1, base))


def chars_len(chars: Union[str, int]) -> int:
    """Calc length of chars"""
    chars = str(chars)
    res = 0
    if isinstance(chars, int):
        chars = str(chars)
    for ch in chars:
        if ord(ch) > 255:
            res += 2
        else:
            res += 1
    return res


def max_key_len(d: dict, offset: int = 0, use_chars_len: bool = True) -> int:
    """Get max string length of dict keys"""
    if not d:
        return 0
    len_func = chars_len if use_chars_len else len
    return max([len_func(str(k)) + offset for k in d.keys()])


def is_str_float(s: str) -> bool:
    return s.replace(".", "", 1).isdigit() and s.count(".") < 2


def to_digits(s: Union[str, int, float], precision: int = None) -> float:
    if isinstance(s, (int, float)):
        return s
    else:  # str
        try:
            return int(s)
        except:
            if precision is None:
                return float(s)
            else:
                return round(float(s), precision)


def get_by_threshold(
    d: Union[dict, list],
    threshold: Union[int, float],
    direction: Literal["upper_bound", "lower_bound"] = "lower_bound",
    target: Literal["key", "value"] = "key",
    is_to_digits: bool = True,
    digits_precision: int = None,
    is_sort: bool = True,
) -> tuple:
    """Get item from dict, which compares on target (key/value) against threshold with direction:
        - threshold as `upper_bound`: get maximum item that is lower than threshold
        - threshold as `lower_bound`: get minimum item that is higher than threshold

    params:
        - d (dict): keys should be int, float, or string (need to convert to float)
        - threshold: numeric threshold to compare against.
        - direction: "upper_bound" or "lower_bound"
        - target: "key" or "value", which to sort and compare against the threshold.

    return:
        - tuple (key, value); if no such entry exists, return (None, None)
    """
    # unify to list of tuples
    if isinstance(d, dict):
        items = list(d.items())
    else:
        items = d
    # unify to number (int/float)
    if is_to_digits:
        if target == "value":
            items = [(k, to_digits(v, digits_precision)) for k, v in items]
        else:
            items = [(to_digits(k, digits_precision), v) for k, v in items]
    # sort, asc, from low to high
    if is_sort:
        if target == "key":
            sorted_items = sorted(items, key=lambda item: item[0])
        else:  # target == "value"
            sorted_items = sorted(items, key=lambda item: item[1])
    # Based on direction, filter the sorted items.
    last_item = (None, None)
    if direction == "upper_bound":
        for item in sorted_items:
            if target == "key":
                if item[0] <= threshold:
                    last_item = item
                else:
                    break
            else:  # target == "value"
                if item[1] <= threshold:
                    last_item = item
                else:
                    break
    else:  # direction == "lower_bound"
        for item in reversed(sorted_items):
            if target == "key":
                if item[0] >= threshold:
                    last_item = item
                else:
                    break
            else:  # target == "value"
                if item[1] >= threshold:
                    last_item = item
                else:
                    break
    return last_item



from .matches import match_paths
from .types import PathType, PathsType
from .paths import norm_path
from .logs import TCLogger
from .confirms import confirm_input

from shutil import copyfile, rmtree
from pathlib import Path

logger = TCLogger(__name__)


def copy_file(src: PathType, dst: PathType):
    src = Path(src)
    dst = Path(dst)
    if not dst.parent.exists():
        dst.parent.mkdir(parents=True, exist_ok=True)
    copyfile(src, dst)


def copy_file_relative(src: PathType, src_root: PathType, dst_root: PathType):
    src = Path(src)
    src_root = Path(src_root)
    dst_root = Path(dst_root)
    dst = dst_root / src.relative_to(src_root)
    copy_file(src, dst)


def copy_folder(
    src_root: PathType,
    dst_root: PathType,
    includes: PathsType = None,
    excludes: PathsType = None,
    use_gitignore: bool = True,
    confirm_before_copy: bool = True,
    remove_existing: bool = True,
    confirm_before_remove: bool = True,
):
    src_root = norm_path(src_root)
    dst_root = norm_path(dst_root)

    logger.note(f"> Copying folder:")
    logger.file(f"  * src: {src_root}")
    logger.file(f"  * dst: {dst_root}")

    if not src_root.exists():
        logger.warn(f"× src not found: {src_root}")
        return

    if confirm_before_copy:
        if not confirm_input("copy", op_name="copy folder"):
            logger.warn("× copy_folder cancelled by user")
            return

    if dst_root.exists():
        logger.warn(f"  × dst exists: {dst_root}")
        if not confirm_input("force", op_name="force copy to existing folder"):
            logger.warn("× copy_folder cancelled as dst exists")
            return
        if remove_existing:
            if confirm_before_remove:
                if not confirm_input(
                    "remove", op_name="remove existing dst folder before copy"
                ):
                    logger.warn("× remove_folder cancelled by user")
                    return
            logger.warn(f"  * remove existing folder: {dst_root}")
            rmtree(dst_root)

    src_files = match_paths(
        src_root,
        includes=includes,
        excludes=excludes,
        use_gitignore=use_gitignore,
    ).get("includes", [])

    for src_file in src_files:
        copy_file_relative(src=src_file, src_root=src_root, dst_root=dst_root)



import threading

from pathlib import Path
from typing import Literal, Union

from .times import get_now_str


MSG_PREFIXES = {"note": ">", "error": "×", "success": "√"}


class FileLogger:
    def __init__(self, log_path: Union[str, Path], lock: threading.Lock = None):
        if not isinstance(log_path, Path):
            log_path = Path(log_path)
        self.log_path = log_path
        if not self.log_path.parent.exists():
            self.log_path.parent.mkdir(parents=True, exist_ok=True)
        self.lock = lock or threading.Lock()

    def log(
        self,
        msg: str,
        msg_type: Literal["note", "error", "success"] = None,
        prefix: str = None,
        add_now: bool = True,
    ):
        if prefix:
            prefix_str = f"{prefix} "
        elif msg_type:
            prefix_str = MSG_PREFIXES.get(msg_type, "*") + " "
        else:
            prefix_str = ""

        if add_now:
            line = f"{prefix_str}[{get_now_str()}] {msg}\n"
        else:
            line = f"{prefix_str}{msg}\n"
        with self.lock:
            with open(self.log_path, "a") as f:
                f.write(line)
